'''Cognite API SDK

This package has been created to ensure excellent excellent user experience for data scientists using the Cognite
Data Platform (CDP).
'''

# RELEASE NAMING CONVENTIONS
#
# Format:
#
# MAJOR.MINOR[.MICRO]
# The major and minor version numbers should mirror the Cognite API. Micro releases are dedicated to bug fixes,
# improvements, and additions.
#


__all__ = ['assets', 'config', 'data_objects', 'tagmatching', 'timeseries', 'raw', 'preprocessing']
__version__ = '0.4.29'
